//
//  AppDelegate.swift
//  MotorPositionController
//
//  Created by Phidgets on 2018-02-23.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = MotorPositionController()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var eventBox: NSBox!
    @IBOutlet weak var motorPositionLabel: NSTextField!
    @IBOutlet weak var motorDutyCycleLabel: NSTextField!
    @IBOutlet weak var targetPositionBox: NSBox!
    @IBOutlet weak var targetPositionText: NSTextField!
    @IBOutlet weak var targetPositionStepper: NSStepper!
    @IBOutlet weak var incrementLabel: NSTextField!
    @IBOutlet weak var configurationBox: NSBox!
    @IBOutlet weak var kpText: NSTextField!
    @IBOutlet weak var kpStepper: NSStepper!
    @IBOutlet weak var kiText: NSTextField!
    @IBOutlet weak var kiStepper: NSStepper!
    @IBOutlet weak var kdText: NSTextField!
    @IBOutlet weak var kdStepper: NSStepper!
    @IBOutlet weak var accelerationText: NSTextField!
    @IBOutlet weak var accelerationStepper: NSStepper!
    @IBOutlet weak var velocityText: NSTextField!
    @IBOutlet weak var velocityStepper: NSStepper!
    @IBOutlet weak var rescaleText: NSTextField!
    @IBOutlet weak var deadbandText: NSTextField!
    @IBOutlet weak var engageButton: NSButton!
    @IBOutlet weak var otherMotorSettingsBox: NSBox!
    @IBOutlet weak var currentLimitBox: NSBox!
    @IBOutlet weak var currentLimitText: NSTextField!
    @IBOutlet weak var currentLimitStepper: NSStepper!
    @IBOutlet weak var currentRegulatorGainBox: NSBox!
    @IBOutlet weak var currentRegulatorGainText: NSTextField!
    @IBOutlet weak var currentRegulatorGainStepper: NSStepper!
    @IBOutlet weak var stallVelocityBox: NSBox!
    @IBOutlet weak var stallVelocityText: NSTextField!
    @IBOutlet weak var stallVelocityStepper: NSStepper!
    @IBOutlet weak var fanModeBox: NSBox!
    @IBOutlet weak var fanModePopup: NSPopUpButton!
    @IBOutlet weak var encoderIOModeBox: NSBox!
    @IBOutlet weak var encoderIOModePopup: NSPopUpButton!
    
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0

    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            frame.size.width = self.window.minSize.width
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.positionChange.addHandler(positionchange_handler)
            let _ = ch.dutyCycleUpdate.addHandler(dutycycleupdate_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.positionChange.removeAllHandlers()
            ch.dutyCycleUpdate.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! MotorPositionController
            let deviceID = try attachedDevice.getDeviceID()
            var currentLimitHidden:Bool = true
            var currentRegulatorHidden:Bool = true
            var fanModeHidden:Bool = true
            var ioModeHidden:Bool = true
            var stallVelocityHidden:Bool = true
            
            if(deviceID == DeviceID.PN_DCC1000){
                currentLimitHidden = false
                currentRegulatorHidden = false
                ioModeHidden = false
                fanModeHidden = false
            }
            /*else if(deviceID == DeviceID.PN_DCC1001){
                 motorSettings.currentLimit = try ch.getCurrentLimit()
                 currentLimitHidden = false
            }*/
            else if(deviceID == DeviceID.PN_DCC1100){
                stallVelocityHidden = false
            }
            
            self.updateGUI()
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.eventBox.isHidden = false
                self.targetPositionBox.isHidden = false
                self.configurationBox.isHidden = false
                self.currentLimitBox.isHidden = currentLimitHidden
                self.currentRegulatorGainBox.isHidden = currentRegulatorHidden
                self.fanModeBox.isHidden = fanModeHidden
                self.encoderIOModeBox.isHidden = ioModeHidden
                self.stallVelocityBox.isHidden = stallVelocityHidden
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                frame.size.width = self.window.maxSize.width
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.eventBox.isHidden = true
            self.targetPositionBox.isHidden = true
            self.configurationBox.isHidden = true
            
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            frame.size.width = self.window.minSize.width
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func positionchange_handler(sender: MotorPositionController, position: Double){
        DispatchQueue.main.async {
            self.motorPositionLabel.stringValue = String(format:"%.2f", position)
        }
    }
    
    func dutycycleupdate_handler(sender: MotorPositionController, dutyCycle: Double){
        DispatchQueue.main.async {
            self.motorDutyCycleLabel.stringValue = String(format:"%.2f", dutyCycle)
        }
    }
    
    //MARK: GUI Controls
    @IBAction func setTargetPosition(_ sender: Any) {
        do{
            if let value = Double(targetPositionText.stringValue){
                try ch.setTargetPosition(value)
            }
            else{
                targetPositionText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func incrementTargetPosition(_ sender: Any) {
        do{
            try ch.setTargetPosition(targetPositionStepper.doubleValue)
            targetPositionText.stringValue = String(format:"%.1f",targetPositionStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setIncrement(_ sender: Any) {
        if let value = Double(incrementLabel.stringValue){
            targetPositionStepper.increment = value
        }
        else{
            incrementLabel.stringValue = "Invalid"
        }
    }
    
    @IBAction func zeroPosition(_ sender: Any) {
        do{
            let position = try ch.getPosition()
            try ch.addPositionOffset(positionOffset: -1 * position)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setKp(_ sender: Any) {
        do{
            if let value = Double(kpText.stringValue){
                try ch.setKp(value)
            }
            else{
                kpText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementKp(_ sender: Any) {
        do{
            try ch.setKp(kpStepper.doubleValue)
            kpText.stringValue = String(format:"%.1f",kpStepper.doubleValue)
            kpStepper.increment = updateLimit(n: kpStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setKi(_ sender: Any) {
        do{
            if let value = Double(kiText.stringValue){
                try ch.setKi(value)
            }
            else{
                kiText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementKi(_ sender: Any) {
        do{
            try ch.setKi(kiStepper.doubleValue)
            kiText.stringValue = String(format:"%.1f",kiStepper.doubleValue)
            kiStepper.increment = updateLimit(n: kiStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setKd(_ sender: Any) {
        do{
            if let value = Double(kdText.stringValue){
                try ch.setKd(value)
            }
            else{
                kdText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementKd(_ sender: Any) {
        do{
            try ch.setKd(kiStepper.doubleValue)
            kdText.stringValue = String(format:"%.1f",kdStepper.doubleValue)
            kdStepper.increment = updateLimit(n: kdStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setAcceleration(_ sender: Any) {
        do{
            if let value = Double(accelerationText.stringValue){
                try ch.setAcceleration(value)
            }
            else{
                accelerationText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementAcceleration(_ sender: Any) {
        do{
            try ch.setAcceleration(accelerationStepper.doubleValue)
            accelerationText.stringValue = String(format:"%.1f",accelerationStepper.doubleValue)
            accelerationStepper.increment = updateLimit(n: accelerationStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setVelocityLimit(_ sender: Any) {
        do{
            if let value = Double(velocityText.stringValue){
                try ch.setVelocityLimit(value)
            }
            else{
                velocityText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementVelocityLimit(_ sender: Any) {
        do{
            try ch.setVelocityLimit(velocityStepper.doubleValue)
            velocityText.stringValue = String(format:"%.1f",velocityStepper.doubleValue)
            velocityStepper.increment = updateLimit(n: velocityStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setRescaleFactor(_ sender: Any) {
        do{
            if let value = Double(rescaleText.stringValue){
                try ch.setRescaleFactor(value)
                updateRescaleProperties()
            }
            else{
                rescaleText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setDeadband(_ sender: Any) {
        do{
            if let value = Double(deadbandText.stringValue){
                try ch.setDeadBand(value)
            }
            else{
                deadbandText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func engageMotor(_ sender: Any) {
        do{
            let button = sender as! NSButton
            let engaged = try ch.getEngaged()
            try ch.setEngaged(!engaged)
            if(engaged){
                button.title = "Engage Motor"
            }
            else{
                button.title = "Disengage Motor"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    
    @IBAction func showOtherSettings(_ sender: Any) {
        let button = sender as! NSButton
        var frame = self.window.frame
        if(button.state == NSControl.StateValue.on){
            frame.size.height = self.window.maxSize.height
        }
        else{
            frame.size.height = self.window.maxSize.height - otherMotorSettingsBox.frame.size.height
        }
        self.window.setFrame(frame, display: true, animate: false)
        self.window.center()
    }
    
    @IBAction func setCurrentLimit(_ sender: Any) {
        do{
            if let value = Double(currentLimitText.stringValue){
                try ch.setCurrentLimit(value)
            }
            else{
                currentLimitText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementCurrentLimit(_ sender: Any) {
        do{
            try ch.setCurrentLimit(currentLimitStepper.doubleValue)
            currentLimitText.stringValue = String(format:"%.1f",currentLimitStepper.doubleValue)
            currentLimitStepper.increment = updateLimit(n: currentLimitStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCurrentRegulatorGain(_ sender: Any) {
        do{
            if let value = Double(currentRegulatorGainText.stringValue){
                try ch.setCurrentRegulatorGain(value)
            }
            else{
                currentRegulatorGainText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementCurrentRegulatorGain(_ sender: Any) {
        do{
            try ch.setCurrentRegulatorGain(currentRegulatorGainStepper.doubleValue)
            currentRegulatorGainText.stringValue = String(format:"%.1f",currentRegulatorGainStepper.doubleValue)
            currentRegulatorGainStepper.increment = updateLimit(n: currentRegulatorGainStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setStallVelocity(_ sender: Any) {
        do{
            if let value = Double(stallVelocityText.stringValue){
                try ch.setStallVelocity(value)
            }
            else{
                stallVelocityText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func incrementStallVelocity(_ sender: Any) {
        do{
            try ch.setStallVelocity(stallVelocityStepper.doubleValue)
            stallVelocityText.stringValue = String(format:"%.1f",stallVelocityStepper.doubleValue)
            stallVelocityStepper.increment = updateLimit(n: stallVelocityStepper.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setFanMode(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setFanMode(FanMode.on)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setFanMode(FanMode.off)
            }
            else{
                try ch.setFanMode(FanMode.auto)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setEncoderIOMode(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setIOMode(EncoderIOMode.pushPull)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setIOMode(EncoderIOMode.lineDriver_2K2)
            }
            else if(popup.indexOfSelectedItem == 2){
                try ch.setIOMode(EncoderIOMode.lineDriver_10K)
            }
            else if(popup.indexOfSelectedItem == 3){
                try ch.setIOMode(EncoderIOMode.openCollector_2K2)
            }
            else{
                try ch.setIOMode(EncoderIOMode.openCollector_10K)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Misc
    func updateLimit(n:Double) -> Double{
        let x = fabs(n)
        if(x != 0 && x > 1){
            return pow(10,floor(log10(x)))/2
        }
        else{
            return 0.5
        }
    }
    
    func updateRescaleProperties(){
        do{
            let deviceID = try ch.getDeviceID()
            let minAcceleration = try ch.getMinAcceleration()
            let maxAcceleration = try ch.getMaxAcceleration()
            let acceleration = try ch.getAcceleration()
            let minVelocityLimit = try ch.getMinVelocityLimit()
            let maxVelocityLimit = try ch.getMaxVelocityLimit()
            let velocityLimit = try ch.getVelocityLimit()
            let deadband = try ch.getDeadBand()
            let minPosition = try ch.getMinPosition()
            let maxPosition = try ch.getMaxPosition()
            var minStallVelocity:Double?, maxStallVelocity:Double?, stallVelocity:Double?
            if(deviceID == DeviceID.PN_DCC1100){
                minStallVelocity = try ch.getMinStallVelocity()
                maxStallVelocity = try ch.getMaxStallVelocity()
                stallVelocity = try ch.getStallVelocity()
            }
            
            DispatchQueue.main.async {
                self.accelerationText.stringValue = String(format:"%.2f",acceleration)
                self.velocityText.stringValue = String(format:"%.2f", velocityLimit)
                self.deadbandText.stringValue = String(format:"%.2f",deadband)
                if(deviceID == DeviceID.PN_DCC1100){
                    self.stallVelocityStepper.minValue = minStallVelocity!
                    self.stallVelocityStepper.maxValue = maxStallVelocity!
                    self.stallVelocityText.stringValue = String(format:"%.2f", stallVelocity!)
                }
                self.accelerationStepper.increment = self.updateLimit(n: acceleration)
                self.accelerationStepper.doubleValue = acceleration
                self.accelerationStepper.minValue = minAcceleration
                self.accelerationStepper.maxValue = maxAcceleration
                self.velocityStepper.increment = self.updateLimit(n: velocityLimit)
                self.velocityStepper.doubleValue = velocityLimit
                self.velocityStepper.minValue = minVelocityLimit
                self.velocityStepper.maxValue = maxVelocityLimit
                self.targetPositionStepper.minValue = minPosition
                self.targetPositionStepper.maxValue = maxPosition
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func updateGUI(){
        do{
            let deviceID = try ch.getDeviceID()
            let kp = try ch.getKp()
            let ki = try ch.getKi()
            let kd = try ch.getKd()
            let rescale = try ch.getRescaleFactor()
            let engaged = try ch.getEngaged()
            var minCurrentLimit:Double?, maxCurrentLimit:Double?, currentLimit:Double?
            var minCurrentRegulatorGain:Double?, maxCurrentRegulatorGain:Double?, currentRegulatorGain:Double?
            var ioMode:EncoderIOMode?
            var minStallVelocity:Double?, maxStallVelocity:Double?, stallVelocity:Double?
            var fanMode:FanMode?
            
            if(deviceID == DeviceID.PN_DCC1000 /*|| deviceID == DeviceID.PN_DCC1001*/){
                minCurrentLimit = try ch.getMinCurrentLimit()
                maxCurrentLimit = try ch.getMaxCurrentLimit()
                currentLimit = try ch.getCurrentLimit()
            }
            if(deviceID == DeviceID.PN_DCC1000){
                ioMode = try ch.getIOMode()
                fanMode = try ch.getFanMode()
                minCurrentRegulatorGain = try ch.getMinCurrentRegulatorGain()
                maxCurrentRegulatorGain  = try ch.getMaxCurrentRegulatorGain()
                currentRegulatorGain = try ch.getCurrentRegulatorGain()
            }
            if(deviceID == DeviceID.PN_DCC1100){
                minStallVelocity = try ch.getMinStallVelocity()
                maxStallVelocity = try ch.getMaxStallVelocity()
                stallVelocity = try ch.getStallVelocity()
            }
            
            DispatchQueue.main.async {
                self.engageButton.title = engaged ? "Disengage Motor" : "Engage Motor"
                self.kpText.stringValue = String(format:"%.1f",kp)
                self.kiText.stringValue = String(format:"%.1f",ki)
                self.kdText.stringValue = String(format:"%.1f",kd)
                self.kpStepper.doubleValue = kp
                self.kiStepper.doubleValue = ki
                self.kdStepper.doubleValue = kd
                self.kpStepper.increment = self.updateLimit(n: kp)
                self.kiStepper.increment = self.updateLimit(n: ki)
                self.kdStepper.increment = self.updateLimit(n: kd)
                self.rescaleText.stringValue = String(format:"%.1f",rescale)
                
                if(deviceID == DeviceID.PN_DCC1000 /*|| deviceID == DeviceID.PN_DCC1001*/){
                    self.currentLimitText.stringValue = String(format:"%.2f", currentLimit!)
                    self.currentLimitStepper.doubleValue = currentLimit!
                    self.currentLimitStepper.minValue = minCurrentLimit!
                    self.currentLimitStepper.maxValue = maxCurrentLimit!
                }
                if(deviceID == DeviceID.PN_DCC1000){
                    self.fanModePopup.addItem(withTitle: "On")
                    self.fanModePopup.addItem(withTitle: "Off")
                    self.fanModePopup.addItem(withTitle: "Auto")
                    self.encoderIOModePopup.addItem(withTitle: "Push Pull")
                    self.encoderIOModePopup.addItem(withTitle: "Line Driver 2.2k")
                    self.encoderIOModePopup.addItem(withTitle: "Line Driver 10k")
                    self.encoderIOModePopup.addItem(withTitle: "Open Collector 2.2k")
                    self.encoderIOModePopup.addItem(withTitle: "Open Collector 10k")
                    self.currentRegulatorGainText.stringValue = String(format:"%.2f",currentRegulatorGain!)
                    self.currentRegulatorGainStepper.doubleValue = currentRegulatorGain!
                    self.currentRegulatorGainStepper.minValue = minCurrentRegulatorGain!
                    self.currentRegulatorGainStepper.maxValue = maxCurrentRegulatorGain!
                    self.encoderIOModePopup.selectItem(at: ioMode == EncoderIOMode.pushPull ? 0 : ioMode == EncoderIOMode.lineDriver_2K2 ? 1 : ioMode == EncoderIOMode.lineDriver_10K ? 2 : ioMode == EncoderIOMode.openCollector_2K2 ? 3 : 4)
                    self.fanModePopup.selectItem(at: fanMode == FanMode.on ? 0 : fanMode == FanMode.off ? 1 : 2)
                }
                if(deviceID == DeviceID.PN_DCC1100){
                    self.stallVelocityText.stringValue = String(format:"%.2f", stallVelocity!)
                    self.stallVelocityStepper.doubleValue = stallVelocity!
                    self.stallVelocityStepper.minValue = minStallVelocity!
                    self.stallVelocityStepper.maxValue = maxStallVelocity!
                }
            }
            
            self.updateRescaleProperties()
            
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
}

